#
# Copyright 2014 Anselm Kruis
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

"""
Installer for Stackless Python (Windows)

This distribution adds a Stackless Python executable to an existing
CPython installation.
"""

from __future__ import absolute_import

import argparse
import sysconfig
import sys
import atexit
import tempfile
import tarfile
import types
import warnings
import subprocess
from os.path import join, normpath, dirname, basename, abspath, isabs, pardir, isdir, isfile
from os import remove as os_remove, walk, chmod, getcwd, chdir
from distutils.file_util import copy_file
from distutils.dir_util import copy_tree, mkpath, remove_tree
from distutils.util import get_platform, byte_compile
from distutils.extension import Extension
import distutils.core
from distutils import log
from pkg_resources import resource_filename, cleanup_resources  # @UnresolvedImport

# we need the correct lib directory. The os module works fine
from os import __file__ as os__file__
from ._version import *

INSTALLER_VERSION_STR = ".".join(map(str,STACKLESS_VERSION + INSTALLER_VERSION))

       
# Build the shared library for Linux
EPREFIX="/EXEC_PREFIX-DOES-NOT-EXSIST"
PREFIX="/PREFIX-DOES-NOT-EXSIST"
PYTHONPATH="/PYTHONPATH-DOES-NOT-EXSIST:stackless-packages::plat-linux2:lib-tk:lib-old:/"
MULTIARCH="SLP_INSTALLER_MULTIARCH-DOES-NOT-EXIST"

def _setValueInBytes(bytes, marker, value):
    lMarker = len(marker)
    lValue  = len(value)
    if lMarker < lValue:
        marker += "-" * (lValue - lMarker)
        l = lValue
    else:
        value += "\x00" * (lMarker - lValue)
        l = lMarker
    assert len(marker) == len(value)
    assert len(marker) == l
        
    i = 0
    while True:
        i = bytes.find(marker, i)
        if -1 == i:
            break
        bytes[i:i+l] = value
        bytes[i+l] = 0

        
def build_pylib_linux(src, dstDir, stdArgs):
    with open(src, "rb") as s:
        lib = bytearray(s.read())
    _setValueInBytes(lib, PREFIX, sysconfig.get_config_var("prefix"))
    _setValueInBytes(lib, EPREFIX, sysconfig.get_config_var("exec_prefix"))
    _setValueInBytes(lib, PYTHONPATH, "stackless-packages:" + sysconfig.get_config_var("PYTHONPATH"))
    try:
        ma = sys._multiarch  # @UndefinedVariable
    except AttributeError:
        ma = "\x00"
    _setValueInBytes(lib, MULTIARCH, ma)

    target = join(dstDir, basename(src))
    mkpath(dstDir, **stdArgs)
    log.info("creating library %s", target)
    if not stdArgs['dry_run']:
        with open(target, "wb") as t:
            t.write(lib)

def remove_file(path, verbose=1, dry_run=0):
    if verbose >= 1:
        log.info("removing '%s'", path)
    if not dry_run:
        os_remove(path)

def compile_missing_modules(source_dir, missing_names, install_dir, debug, stdArgs):
    # create the pyconfig.h
    pyconfig = sysconfig.get_config_h_filename()
    copy_file(pyconfig, join(source_dir, "pyconfig.h"), verbose=stdArgs['verbose'])
    # create 'Modules/Setup', 'Modules/Setup.local'
    open(join(source_dir, "Modules", "Setup"), "wb").close()
    open(join(source_dir, "Modules", "Setup.local"), "wb").close()
    
    # source the setup script
    setup = types.ModuleType("setup")
    with open(join(source_dir, "setup.py"), "rU") as fp:
        # we need a function to avoid 
        # SyntaxError: unqualified exec is not allowed in function 'compile_missing_modules' it contains a nested function with free variables
        def exec_setup(fp, d):
            exec(fp, d, d)
        exec_setup(fp, setup.__dict__)
    
    class BuildExt(setup.PyBuildExt):
        def detect_modules(self):
            missing = setup.PyBuildExt.detect_modules(self)

            # now filter the modules
            extensions = [ext for ext in self.extensions if ext.name in missing_names]
            self.extensions = extensions
            missing = [name for name in missing if name in missing_names]
            
            # disable the import test. The host python can#t import some modules
            setup.cross_compiling = True  # must be set late
            
            return missing
            
        def build_extensions(self):
            cwd = getcwd()
            old_src = sysconfig.get_config_var('srcdir')
            sysconfig.get_config_vars()['srcdir'] = source_dir

            chdir(source_dir)
            old_builtin_module_names = sys.builtin_module_names
            sys.builtin_module_names = tuple(name for name in sys.builtin_module_names if name not in missing_names)
            try:
                setup.PyBuildExt.build_extensions(self)
            finally:
                sysconfig.get_config_vars()['srcdir'] = old_src
                sys.builtin_module_names = old_builtin_module_names
                chdir(cwd)
                
        def build_extension(self, ext):
            # work around the missing cross compile support in earlier versions
            # of the python setup.py script
            try:
                return setup.PyBuildExt.build_extension(self, ext)
            except (OSError, IOError):
                if not stdArgs['dry_run']:
                    raise

    cmd_name = 'build_ext'
    with warnings.catch_warnings():
        build_temp = join(source_dir, "build_temp")
        mkpath(build_temp, **stdArgs)
        warnings.simplefilter("ignore",category=DeprecationWarning)
        if debug:
            distutils.core.DEBUG = True
        dist = distutils.core.setup(# PyPI Metadata (PEP 301)
                             name = "Python",
                             version = sys.version.split()[0],
                             # Build info
                             cmdclass = {cmd_name:BuildExt},
                             # The struct module is defined here, because build_ext won't be
                             # called unless there's at least one extension module defined.
                             ext_modules=[Extension('_struct', ['_struct.c'])],
                             script_args=[cmd_name],
                             options=dict(build_ext=dict(build_lib=install_dir, build_temp=build_temp)),
                             scripts = [],  # the code in setup.PyBuildExt assumes scripts beeing itarable
                             **stdArgs
                             )
    cmd = dist.get_command_obj(cmd_name, create=False)
    if cmd.failed:
        return "Failed to compile some modules: " + ", ".join(cmd.failed)
    return None


def build_missing_modules(modules_src_archive, missing_names, install_dir, debug, stdArgs):
    basedir = tempfile.mkdtemp(prefix="stackless_installer_")
    try:
        with tarfile.open(modules_src_archive, "r",
                          debug=stdArgs['verbose'] - 1 if stdArgs['verbose'] >= 1 else 0) as tar:
            for tarinfo in tar.getmembers():
                if (not (tarinfo.isfile() or tarinfo.isdir()) or 
                    isabs(tarinfo.name) or normpath(tarinfo.name).startswith(pardir)):
                    # bogous entry
                    continue
                tar.extract(tarinfo, path=basedir)
        # now we have the extracted source
        return compile_missing_modules(basedir, missing_names, install_dir, debug, stdArgs)

    finally:
        remove_tree(basedir, verbose=stdArgs['verbose'], dry_run=stdArgs['dry_run'] or debug)


def test_modules(python, module_names, verbose=1, dry_run=0):
    import_failed = []
    for name in module_names:
        if verbose >= 1:
            log.info("test import of module '%s'", name)
        try:
            if not dry_run:
                subprocess.check_call([python, "-S", "-E", "-c", 'import {}'.format(name)])
        except subprocess.CalledProcessError:
            import_failed.append(name)
    if import_failed:
        return "Error: some modules can't be imported: " + ", ".join(import_failed)
    return None


def main(args, default="install"):
    
    parser = argparse.ArgumentParser(description='Install or uninstall stackless python')
    parser.add_argument('--verbose', '-v', dest='verbose', action='append_const', const=True, help="run verbosely (default)")
    parser.add_argument('--quiet', '-q', dest='verbose', action='store_const', const=[], default=argparse.SUPPRESS, help="run quietly (turns verbosity off)")
    parser.add_argument('--dry-run', '-n', dest='dry_run', action='store_true', help="don't actually do anything")
    parser.add_argument('--debug', dest='debug', action='store_true', help="don't clean up")
    parser.add_argument('command', choices=['install', 'uninstall'], action='store', nargs='?', default=default)
    
    parsed = parser.parse_args(args=args)
    print_success = parsed.verbose is None or len(parsed.verbose)
    parsed.verbose = 0 if parsed.verbose is None else len(parsed.verbose)
    log.set_verbosity(parsed.verbose)
    
    if get_platform() != PLATFORM:
        raise RuntimeError("This package is for %s only. It does not work on %s" % (PLATFORM, get_platform()))
    currentUcs = 4 if sys.maxunicode >> 16 else 2
    if currentUcs != UCS:
        raise RuntimeError("This package requires ucs%d but your Python interpreter provides ucs%d" % (UCS, currentUcs))

    # standard arguments of various functions from distutils    
    stdArgs = dict(verbose=parsed.verbose, dry_run=parsed.dry_run)
    
    arch = PLATFORM
    assembly="lib.stackless"
    slPython = 'slpython%s.%s' % STACKLESS_VERSION[:2]
    pybindest = dirname(abspath(sys.executable))
    pylibdest = join(pybindest, assembly)
    pybindest = join(pybindest, slPython)
    stacklessPackages = join(dirname(abspath(os__file__)), "stackless-packages")

    if parsed.command not in ('uninstall', 'install'):
        raise ValueError("Unsupported command: " + parsed.command)

    # remove the executable and the shared library
    if isfile(pybindest):
        remove_file(pybindest, **stdArgs)
    if isdir(pylibdest):
        remove_tree(pylibdest, **stdArgs)
    # remove the lib
    if isdir(stacklessPackages):
        remove_tree(stacklessPackages, **stdArgs)
        
    if parsed.command == 'install':
        # get the binary resources
        atexit.register(cleanup_resources)
        src = resource_filename(__name__, arch)
        with open(join(src, "slp_built_in_modules.txt"), "rU") as fp:
            slp_builtin_module_names = [line.strip() for line in fp]
        modules_src_archive = join(src, "modules_src.tar.gz")
        
        # copy the executable
        pyexe = join(src, slPython)
        copy_file(pyexe, pybindest, **stdArgs)
        if not stdArgs['dry_run']:
            chmod(pybindest, 0755)
        # copy and modify the lib
        pylib = join(src, assembly, "libpython%s.%s.so.1.0sl" % STACKLESS_VERSION[:2] )
        build_pylib_linux(pylib, pylibdest, stdArgs)
        
        # get the binary resources
        src = resource_filename(__name__, 'Lib')
        copy_tree(src, stacklessPackages, **stdArgs)
        
        # byte compile sources
        pyFiles = [] 
        for root, dirs, files in walk(stacklessPackages):  # @UnusedVariable
            pyFiles.extend(join(root, f) for f in files if f.endswith('.py'))
        byte_compile(pyFiles, optimize=False, prefix=stacklessPackages, force=1, **stdArgs)
        
        # build missing extension modules
        missing_names = []
        for name in sys.builtin_module_names:
            if name not in slp_builtin_module_names:
                missing_names.append(name)
        if missing_names:
            errors = []
            # some modules are missing
            error = build_missing_modules(modules_src_archive, missing_names, stacklessPackages, parsed.debug, stdArgs)
            if error:
                errors.append(error)
            error = test_modules(pybindest, missing_names, **stdArgs)
            if error:
                errors.append(error)
            if errors:
                return "\n".join(errors)

    if print_success:
        print "Stackless installer:", parsed.command, "successfully completed"


def main_install():
    return main(args=sys.argv[1:], default='install')
def main_uninstall():
    return main(args=sys.argv[1:], default='uninstall')

if __name__ == '__main__':
    sys.exit(main(args=sys.argv[1:]))
